"""
utils
"""
from .time_conversions import utc_to_lst, normalize_time_to_decimal, LST_DAY_DEC
from .normalize_coordinates import normalize_coordinates
from .normalize_date import normalize_yyyymmdd_to_datetime, normalize_datetime

__all__ = [
    "normalize_coordinates",
    "utc_to_lst",
    "normalize_time_to_decimal",
    "LST_DAY_DEC",
    "normalize_yyyymmdd_to_datetime",
    "normalize_datetime",
]

# Automatically added by katversion
__version__ = '1.8.3'



from ...observation import Observation
from ...lstindex import LSTIntervalType as I
import math
from typing import Tuple
from lstpressure.perf import track_total_runtime


def convert_to_hours(x):
    try:
        result = round(
            (int(x.split(":")[0]) * 3600 + int(x.split(":")[1]) * 60) / 3600, 2
        )
        return result
    except:
        return None


@track_total_runtime
def calculate_constraints(row) -> list[I]:
    constraints = []
    if row["night_obs"] == "Yes":
        constraints.append(I.NIGHT)
    elif row["avoid_sunrise_sunset"] == "Yes":
        constraints.append(I.SUNRISE_SUNSET)
        constraints.append(I.SUNSET_SUNRISE)
        constraints.append(I.DAY)

    # If neither night_obs nor avoid_sunrise_sunset was selected
    if not constraints:
        constraints.append(I.ALL_DAY)
        constraints.append(I.DAY)

    return constraints


@track_total_runtime
def split_dates(dict):
    if not len(dict.keys()):
        return None

    # Dictionary to store the start and end dates
    dates = {}

    # Iterate through the dictionary items
    for key, value in dict.items():
        # Check if the key ends with 'start_date' or 'end_date'
        if "start_date" in key or "end_date" in key:
            # Extract the base key by removing the suffix '_start_date' or '_end_date'
            base_key = key.replace("_start_date", "").replace("_end_date", "")
            if base_key not in dates:
                dates[base_key] = {}
            # Assign the value to the respective part of the tuple (start or end)
            if "start_date" in key:
                dates[base_key]["start"] = value
            elif "end_date" in key:
                dates[base_key]["end"] = value

    # Convert the dictionary to a list of tuples
    result = []
    for key in dates:
        if "start" in dates[key] and "end" in dates[key]:
            result.append((dates[key]["start"], dates[key]["end"]))

    return result


@track_total_runtime
def calc_tags(dict):
    if not len(dict.keys()):
        return None

    # 'tag' could be suffix for a number of fields. make sure we are only working with the intended field
    dict = {
        key: val for key, val in dict.items() if key == "tag" or key.startswith("tag_")
    }
    return [val for _, val in dict.items()]


@track_total_runtime
def wind_normalized_field(row, field):
    return {
        key: str(value)
        for key, value in row.items()
        if key.startswith(field) and str(value) != "nan"
    }


@track_total_runtime
def calculate_observations(dataFrame, observation_filter) -> Tuple[Observation]:
    try:
        # Convert durations to hours and round off
        dataFrame["duration_hours"] = dataFrame["simulated_duration"].apply(
            lambda x: round(0 if math.isnan(x) else x / 3600, 2)
        )

        # Convert LST start and end times to floating point hours
        dataFrame["lst_start_hours"] = dataFrame["lst_start"].apply(convert_to_hours)
        dataFrame["lst_end_hours"] = dataFrame["lst_start_end"].apply(convert_to_hours)

        # Build observations using list comprehension for efficiency
        observations = [
            Observation(
                id=row["id"],
                lst_window_start=row["lst_start_hours"],
                lst_window_end=row["lst_end_hours"],
                utc_constraints=constraint,
                duration=row["duration_hours"],
                tags=calc_tags(wind_normalized_field(row, "tag")),
                preferred_dates=split_dates(
                    wind_normalized_field(row, "preferred_dates")
                ),
                avoid_dates=split_dates(wind_normalized_field(row, "avoid_dates")),
                proposal_id=row["proposal_id"],
            )
            for _, row in dataFrame.iterrows()
            if row["lst_end_hours"] - row["lst_start_hours"] > 0
            for constraint in calculate_constraints(row)
        ]

        return tuple(filter(observation_filter, observations))
    except Exception as e:
        raise ValueError("Error parsing CSV - please check input.", e)



from abc import ABC, abstractmethod
from typing import Self


class AppInterface(ABC):
    @property
    @staticmethod
    def id():
        pass

    @property
    @staticmethod
    def usage():
        pass

    @property
    @staticmethod
    def description():
        pass

    def __init__(self, parser) -> None:
        self.parser = parser
        self.build()

    @abstractmethod
    def build(self) -> Self:
        pass

    @abstractmethod
    def parse(self) -> Self:
        pass

    @abstractmethod
    def exe(self) -> None:
        pass



from ..AppInterface import AppInterface
from typing import Self
import os
import subprocess
from lstpressure.logger import error
import sys
import csv
from io import StringIO


SQL_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "sql")
existing_reports = [
    f.replace(".sql", "")
    for f in os.listdir(SQL_FOLDER_PATH)
    if os.path.isfile(os.path.join(SQL_FOLDER_PATH, f))
]
quoted_existing_reports = ", ".join([f'"{r}"' for r in existing_reports])


class Aggregate(AppInterface):
    id = "aggregate"
    usage = "lstpressure aggregate -h"
    description = "Aggregate observables using built-in reports or bespoke SQL commands"

    def __init__(self, parser) -> None:
        super().__init__(parser)

    def build(self) -> Self:
        self.parser.add_argument(
            "--input",
            required=False,
            type=str,
            metavar="",
            help="Path to lst-pressure output csv file (if not provided, stdin is used instead)",
        )
        self.parser.add_argument(
            "--query",
            required=True,
            type=str,
            metavar="",
            help=f'SQL string to apply to input CSV (i.e. "select * from stdin"), or apply a built-in aggregation: {quoted_existing_reports}. By default the SQLite engine is used, and you can enable the DuckDB engine by including "duckdb" either in a filename or as a comment somewhere in the query text',
        )
        self.parser.add_argument(
            "--pretty",
            required=False,
            default=False,
            action="store_true",
            help=f"Format CSV output using the csvlook tool (included in this CLI)",
        )
        self.parser.add_argument(
            "--no-trunc",
            required=False,
            default=False,
            action="store_true",
            help=f"When using --pretty, long values are truncated by default. Disable this behaviour via this flag",
        )
        self.parser.add_argument(
            "--echo",
            required=False,
            default=False,
            action="store_true",
            help=f"Echo the aggregation query back - useful for evaluating .sql file contents that would otherwise be opaque. For example: --echo --query {existing_reports[0]}",
        )

        return self

    def parse(self, args) -> Self:
        self.args = args
        self.echo = args.echo
        self.input = args.input
        self.query = parse_sql_query(args.query)
        self.pretty = args.pretty
        self.no_trunc = args.no_trunc if args.no_trunc else False

        if self.no_trunc:
            if not self.pretty:
                error("--no-trunc is only applicable when --pretty flag is used")
                exit(1)

        if self.echo and self.pretty:
            error("Don't use --pretty and --echo together.")
            exit(1)

        return self

    def exe(self) -> None:
        if self.echo:
            print(self.query)
            exit()

        if not self.input:
            if not sys.stdin.isatty():
                input_csv = sys.stdin.read()
            else:
                error(
                    "No input provided. Please either provide a filepath (--input), or pipe a CSV-formatted string from stdin"
                )
                exit(1)
        else:
            with open(self.input, "r") as file:
                input_csv = file.read()

        result = execute_csvsql(input_csv, self.query, self.args.query)
        result = execute_csvlook(result, self.no_trunc) if self.pretty else result

        # Print result to stdout
        print(result)
        exit()


def parse_sql_query(str_input):
    # A .sql file
    if os.path.isfile(str_input):
        with open(str_input, "r") as file:
            return file.read().strip()

    # Try to find the SQL file if it's included in the tool
    path = os.path.join(SQL_FOLDER_PATH, f"{str_input}")

    # Name of view without .sql
    if os.path.isfile(path):
        with open(path, "r") as file:
            return file.read().strip()

    # Name of view with .sql
    if os.path.isfile(path + ".sql"):
        with open(path + ".sql", "r") as file:
            return file.read().strip()

    # If not a file, assume it's a SQL string
    return str_input


def truncate_cols_to_max_chars(max_chars, line):
    # Use StringIO to simulate file I/O which csv.reader expects
    input_line = StringIO(line)
    output_line = StringIO()
    reader = csv.reader(input_line)
    writer = csv.writer(output_line)

    # Read the line with csv.reader which handles commas, quotes correctly
    for row in reader:
        # Truncate each field to max_chars and append ellipsis if truncated
        truncated_row = [
            (field[: max_chars - 3] + " ...") if len(field) > max_chars else field
            for field in row
        ]
        # Write the truncated row to the output_line
        writer.writerow(truncated_row)

    # Get the content of output_line and remove the trailing newline
    return output_line.getvalue().strip()


def execute_csvlook(input_content, no_trunc=False):
    csvlook_opts = ["-I", "--null-value", "0"]
    command = ["csvlook", *csvlook_opts]

    try:
        # Start the subprocess with pipes for stdin and stdout
        with subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        ) as process:
            # Check if input is string and split into lines if necessary
            if isinstance(input_content, str):
                input_content = input_content.splitlines()

            # Process each line, writing to stdin and logging
            for line in input_content:
                line = truncate_cols_to_max_chars(18, line) if not no_trunc else line
                process.stdin.write(line + "\n")
                process.stdin.flush()  # Ensure the line is sent to the process

            # Read the output
            output, errors = process.communicate()

            # Check for errors
            if process.returncode != 0:
                print(
                    f"csvlook process returned a non-zero exit code: {process.returncode}"
                )
                print(f"Errors: {errors}")
                return None

            return output

    except Exception as e:
        print(f"An error occurred: {e}")
        raise


def execute_csvsql(input_content, sql, query_name=""):
    duck_opts = (
        [
            "--db",
            "duckdb:///:memory:",
            "--insert",
        ]
        if "duckdb" in query_name or "duckdb" in sql
        else []
    )

    command = ["csvsql", "--query", sql, *duck_opts]

    try:
        # Use subprocess.Popen to run the command
        with subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        ) as process:
            # Pass input content to the subprocess
            output, _ = process.communicate(input=input_content)

            # Wait for the process to finish and get the return code
            return_code = process.wait()

            if return_code == 0:
                return output
            else:
                error(f"csvsql process returned a non-zero exit code: {return_code}")
                error(f"Command output: {output}")
                return None
    except Exception as e:
        raise e


__all__ = ["parse_sql_query", "Aggregate", "execute_csvsql", "execute_csvlook"]

# Automatically added by katversion
__version__ = '1.8.3'



"""
lstpressure.lstindex.LSTInterval
"""

from datetime import datetime
from intervaltree import Interval
from .LSTIntervalType import LSTIntervalType
from ..lstcalendar import LSTCalendarDate
from ..sun import Sun
from ..observation import Observation
from typing import (
    Optional,
    Union,
)  # TODO The new | type (instead of Union doesn't work here)
from ..utils import normalize_yyyymmdd_to_datetime
from lstpressure.perf import track_total_runtime, decorate_all


@track_total_runtime
def normalize_interval(start, end, days=1):
    """
    Normalize an interval's start and end times.

    This function ensures that the end time is greater than or equal to the start time, even if the interval spans multiple days.

    Parameters
    ----------
    start : float
        The start time of the interval (in hours).
    end : float
        The end time of the interval (in hours).
    days : int, optional
        The number of days the interval spans. Defaults to 1.

    Returns
    -------
    tuple
        A tuple containing the normalized start and end times.
    """
    if end < start:
        end += 24
    end += 24 * (days - 1)
    return (start, end)


@decorate_all(track_total_runtime)
class LSTInterval:
    def __init__(
        self,
        start: float,  # lst start
        end: float,  # lst end
        start_utc: datetime | float,
        end_utc: datetime | float,
        parent: Optional[Union[LSTCalendarDate, Observation]] = None,
        dt: Optional[datetime] = None,
        type: Optional[LSTIntervalType] = None,
        sun: Optional[Sun] = None,
        tomorrow_sun: Optional[Sun] = None,
    ) -> None:
        """
        Initialize an LSTInterval object.

        Parameters
        ----------
        start : float
            The start time of the interval (in hours).
        end : float
            The end time of the interval (in hours).
        start_utc : datetime | float
            The start time of the interval in UTC.
        end_utc : datetime | float
            The end time of the interval in UTC.
        parent : LSTCalendarDate | Observation, optional
            The parent object associated with the interval. Defaults to None.
        dt : datetime, optional
            The datetime associated with the interval. Defaults to None.
        type : LSTIntervalType, optional
            The type of the interval. Defaults to None.
        sun : Sun, optional
            The Sun object associated with the interval. Defaults to None.
        tomorrow_sun : Sun, optional
            The Sun object associated with the next day. Defaults to None.
        """
        self._type = type if type else None
        self._dt = normalize_yyyymmdd_to_datetime(dt) if dt else None
        self._sun = sun if sun else None
        self._tomorrow_sun = tomorrow_sun if tomorrow_sun else None
        self._parent = parent if parent else None
        self._interval = Interval(start, end, self)
        self._start_utc = start_utc
        self._end_utc = end_utc

    @property
    def parent(self) -> Union[LSTCalendarDate, Observation]:
        """
        The parent object associated with the interval.

        Returns
        -------
        LSTCalendarDate | Observation
            The parent object.
        """
        return self._parent

    @property
    def interval(self) -> Interval:
        """
        The underlying interval object.

        Returns
        -------
        Interval
            The interval object.
        """
        return self._interval

    @property
    def start(self) -> float:
        """
        The start time of the interval (in hours).

        Returns
        -------
        float
            The start time.
        """
        return self._interval[0]

    @property
    def start_utc(self) -> datetime | float:
        """
        The start time of the interval in UTC.

        Returns
        -------
        datetime | float
            The start time in UTC.
        """
        return self._start_utc

    @property
    def end(self) -> float:
        """
        The end time of the interval (in hours).

        Returns
        -------
        float
            The end time.
        """
        return self._interval[1]

    @property
    def end_utc(self) -> datetime | float:
        """
        The end time of the interval in UTC.

        Returns
        -------
        datetime | float
            The end time in UTC.
        """
        return self._end_utc

    @property
    def type(self) -> Optional[LSTIntervalType]:
        """
        The type of the interval.

        Returns
        -------
        Optional[LSTIntervalType]
            The type of the interval, or None if not specified.
        """
        return self._type

    @property
    def dt(self) -> Optional[datetime]:
        """
        The datetime associated with the interval.

        Returns
        -------
        Optional[datetime]
            The datetime, or None if not specified.
        """
        return self._dt

    @property
    def sun(self) -> Optional[Sun]:
        """
        The Sun object associated with the interval.

        Returns
        -------
        Optional[Sun]
            The Sun object, or None if not specified.
        """
        return self._sun

    @property
    def tomorrow_sun(self) -> Optional[Sun]:
        """
        The Sun object associated with the next day.

        Returns
        -------
        Optional[Sun]
            The Sun object for the next day, or None if not specified.
        """
        return self._tomorrow_sun



from .calculate_observables import calculate_observables
from .calculate_observations import calculate_observations

__all__ = ["calculate_observations", "calculate_observables"]

# Automatically added by katversion
__version__ = '1.8.3'



import pytest
from lstpressure.utils import normalize_coordinates


@pytest.mark.parametrize(
    "latitude, longitude, expected_latitude, expected_longitude",
    [
        ("-30:42:39.8", "21:26:38.0", -30.711055555555554, 21.44388888888889),
        (-30.711055555555554, 21.44388888888889, -30.711055555555554, 21.44388888888889),
    ],
)
def test_normalize_coordinates(latitude, longitude, expected_latitude, expected_longitude):
    lat, long = normalize_coordinates(latitude, longitude)
    assert lat == expected_latitude
    assert long == expected_longitude


def test_normalize_coordinates_mixed_format():
    lat = ["-30:42:39.8", -30.711055555555554]
    long = ["21:26:38.0", 21.44388888888889]

    # Should not be possible to use both DEC and DMS
    with pytest.raises(ValueError):
        normalize_coordinates(lat[0], long[1])
    with pytest.raises(ValueError):
        normalize_coordinates(lat[1], long[0])



import os
from dotenv import load_dotenv
from enum import Enum, auto

current_script_dir = os.path.dirname(os.path.abspath(__file__))
env_path = os.path.join(current_script_dir, "..", ".env")
load_dotenv(env_path)


class LocationProviderType(Enum):
    ASTRAL = auto()
    MEERKAT = auto()


class LogLevel(Enum):
    DEBUG = auto()
    INFO = auto()
    WARN = auto()


class Conf:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Conf, cls).__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        self._log_level = os.environ.get("LOG_LEVEL", LogLevel.INFO)
        self._latitude = os.environ.get("LATITUDE", "-30:42:39.8")
        self._longitude = os.environ.get("LONGITUDE", "21:26:38.0")
        self._py_env = os.environ.get("PY_ENV", "development")

        if os.environ.get("LOC_PROVIDER"):
            self._loc_provider = LocationProviderType[os.environ.get("LOC_PROVIDER").upper()]
        else:
            self._loc_provider = LocationProviderType.MEERKAT

        test_mode_str = os.environ.get("TEST_MODE", "False")
        if test_mode_str and test_mode_str.lower() in ("0", "false"):
            self._test_mode = False
        else:
            self._test_mode = True

    @property
    def LOC_PROVIDER(self) -> bool:
        return self._loc_provider

    @LOC_PROVIDER.setter
    def LOC_PROVIDER(self, value: LocationProviderType) -> None:
        self._loc_provider = value

    @property
    def TEST_MODE(self) -> bool:
        return self._test_mode

    @TEST_MODE.setter
    def TEST_MODE(self, value: bool) -> None:
        self._test_mode = value

    @property
    def LOG_LEVEL(self):
        return self._log_level

    @LOG_LEVEL.setter
    def LOG_LEVEL(self, value):
        self._log_level = value

    @property
    def LATITUDE(self):
        return self._latitude

    @LATITUDE.setter
    def LATITUDE(self, value):
        self._latitude = value

    @property
    def LONGITUDE(self):
        return self._longitude

    @LONGITUDE.setter
    def LONGITUDE(self, value):
        self._longitude = value

    @property
    def PY_ENV(self):
        return self._py_env

    @PY_ENV.setter
    def PY_ENV(self, value):
        self._py_env = value


__all__ = ["Conf", "LogLevel"]

# Automatically added by katversion
__version__ = '1.8.3'



"""
lstpressure.observation.Observation
"""

from typing import Optional, Tuple
from ..lstcalendar.LSTCalendarDate import LSTCalendar, Observable
from ..lstindex import LSTIntervalType, LSTInterval, normalize_interval
from lstpressure.perf import track_total_runtime, decorate_all


@decorate_all(track_total_runtime)
class Observation:
    """
    Represents an observation block with a specified Local Sidereal Time (LST) window and optional UTC constraints.

    Attributes
    ----------
    id : any
        The unique identifier for the observation block.
    lst_window_start : float
        The starting value of the LST window in hours (decimal).
    lst_window_end : float
        The ending value of the LST window in hours (decimal).
    utc_constraints : list[LSTInterval]
        The UTC constraints for the observation block, represented as a list of LSTInterval values.
    proposal_id : Optional[str]
        The proposal ID associated with the observation, if applicable.
    _duration : float
        The duration of the observation in hours (decimal). Defaults to 0 if not specified.
    _tags : list[str]
        A list of tags associated with the observation, used for classification or filtering.
    _preferred_dates : list
        A list of preferred dates for the observation. Defaults to an empty list if not specified.
    _avoid_dates : list
        A list of dates to avoid for the observation. Defaults to an empty list if not specified.
    _cal : Optional[LSTCalendar]
        Reference to the LSTCalendar instance the observation is added to. Defaults to None.
    _intervals : list[LSTInterval]
        A list of LST intervals representing the observation window.
    """

    def __init__(
        self,
        id: any,
        lst_window_start: float,
        lst_window_end: float,
        utc_constraints: LSTIntervalType | list[LSTIntervalType] = None,
        duration: float = None,
        proposal_id: str = None,
        tags: list[str] = None,
        preferred_dates: list = None,
        avoid_dates: list = None,
    ) -> None:
        """
        Initialize an Observation object with specified LST window, optional UTC constraints, duration, and tags.

        Parameters
        ----------
        id : any
            The unique identifier for the observation block.
        lst_window_start : float
            The starting value of the LST window in hours (decimal).
        lst_window_end : float
            The ending value of the LST window in hours (decimal).
        utc_constraints : LSTIntervalType | list[LSTIntervalType], optional
            The UTC constraints as an LSTIntervalType or list of LSTIntervalType values.
        duration : float, optional
            The duration of the observation in hours (decimal). Defaults to 0 if not specified.
        proposal_id : str, optional
            The proposal ID associated with the observation.
        tags : list[str], optional
            A list of tags associated with the observation. Defaults to an empty list if not provided.
        preferred_dates : list, optional
            A list of preferred dates for the observation. Defaults to an empty list if not provided.
        avoid_dates : list, optional
            A list of dates to avoid for the observation. Defaults to an empty list if not provided.
        """
        self.id = id
        self.lst_window_start = lst_window_start
        self.lst_window_end = lst_window_end
        self.utc_constraints = list(
            filter(
                lambda item: item,
                (
                    utc_constraints
                    if utc_constraints and isinstance(utc_constraints, list)
                    else [utc_constraints]
                ),
            )
        )
        self.proposal_id = proposal_id
        self._duration = duration if duration else 0
        self._tags = tags if tags else []
        self._preferred_dates = preferred_dates if preferred_dates else []
        self._avoid_dates = avoid_dates if avoid_dates else []
        self._cal: Optional[LSTCalendar] = None  # Reference to the calendar
        self._intervals = []
        self._intervals.append(
            LSTInterval(
                *normalize_interval(self.lst_window_start, self.lst_window_end),
                None,
                None,
                parent=self,
                type=LSTIntervalType.OBSERVATION_WINDOW,
            )
        )
        if self.lst_window_end < self.lst_window_start:
            self._intervals.append(
                LSTInterval(0, self.lst_window_end, None, None, self)
            )

    @property
    def duration(self) -> float:
        """
        Required observation duration in hours (decimal)
        """
        return self._duration

    @property
    def tags(self) -> list[str]:
        """
        List of tags associated with an observation
        """
        return self._tags

    @property
    def preferred_dates(self) -> list:
        """
        List of preferred dates of an observation
        """
        return self._preferred_dates

    @property
    def avoid_dates(self) -> list:
        """
        List of dates to avoid for an observation
        """
        return self._avoid_dates

    @property
    def intervals(self) -> list[LSTInterval]:
        return self._intervals

    @property
    def calendar(self) -> LSTCalendar:
        if not self._cal:
            raise ValueError("Block has not been added to any LSTCalendar.")
        return self._cal

    @calendar.setter
    def calendar(self, cal: LSTCalendar):
        self._cal = cal

    def __hash__(self) -> int:
        return hash((self.id, self.lst_window_start, self.lst_window_end))

    def observables(
        self, lstcalendar: Optional[LSTCalendar] = None
    ) -> Tuple[Observable]:
        lstcalendar = self._cal if not lstcalendar else lstcalendar

        if not lstcalendar:
            raise ValueError(
                "'lstcalendar' is not specified. To check observability, either associate this observation with an existing LSTCalendar instance or pass an LSTCalendar instance as an argument to this method."
            )

        results = set()

        # Note that overlap() returns a set
        query_result = set()
        for obs_window in self.intervals:
            query_result = query_result | lstcalendar.interval_index.overlap(
                obs_window.interval
            )

            for cal_interval_raw in query_result:
                cal_interval_start, cal_interval_end, cal_interval = cal_interval_raw
                interval_type = cal_interval.type

                if (self.utc_constraints is None or len(self.utc_constraints) == 0) or (
                    len(self.utc_constraints) > 0
                    and interval_type in self.utc_constraints
                ):
                    if obs_window.end > cal_interval_start:
                        if (
                            obs_window.start + self.duration < cal_interval_end
                            or obs_window.end + self.duration < cal_interval_end
                        ):
                            results.add(Observable(cal_interval, self))

        return tuple(results)



import pytest
from lstpressure.lstcalendar import LSTCalendar
from lstpressure.lstindex import LSTIntervalType
from lstpressure.observation import Observation
from lstpressure.conf import Conf, LocationProviderType

conf = Conf()
conf.LOC_PROVIDER = LocationProviderType.ASTRAL

OBSERVATIONS = [
    Observation(
        o.get("id"),
        o.get("lst_window_start"),
        o.get("lst_window_end"),
        o.get("utc_constraints"),
        o.get("duration"),
    )
    for o in [
        {
            "id": "O1",
            "lst_window_start": 6.5,
            "lst_window_end": 10.2,
            "utc_constraints": [LSTIntervalType.ALL_DAY],
            "duration": 1,
        },
        {
            "id": "O2",
            "lst_window_start": 8.5,
            "lst_window_end": 10.2,
            "utc_constraints": [LSTIntervalType.NIGHT],
            "duration": 1,
        },
        {
            "id": "O3",
            "lst_window_start": 21.5,
            "lst_window_end": 20,
            "utc_constraints": [LSTIntervalType.NIGHT],
            "duration": 0.5,
        },
        {
            "id": "O4",
            "lst_window_start": 12,
            "lst_window_end": 22,
            "utc_constraints": [LSTIntervalType.NIGHT],
            "duration": 1.5,
        },
        {
            "id": "O5",
            "lst_window_start": 9,
            "lst_window_end": 10,
            "utc_constraints": [LSTIntervalType.SUNRISE_SUNSET],
            "duration": 0.5,
        },
    ]
]


@pytest.mark.parametrize(
    "dt_range, latlng, expected",
    [
        (["20231101"], ["-30:42:39.8", "21:26:38.0"], ["O1", "O3", "O4", "O5"]),
        (["20231101", "20231101"], ["-30:42:39.8", "21:26:38.0"], ["O1", "O3", "O4", "O5"]),
    ],
)
def test_self_observations(dt_range, latlng, expected: list):
    for d in LSTCalendar(
        *dt_range,
        latitude=latlng[0],
        longitude=latlng[1],
        observations=OBSERVATIONS,
    ).dates:
        observations = set([o.observation.id for o in d.observables()])
        assert sorted(list(observations)) == sorted(expected)



import sys
from lstpressure.conf import Conf, LogLevel
import warnings

conf = Conf()


def debug(*args, **kwargs):
    if conf.LOG_LEVEL == LogLevel.DEBUG:
        print("DEBUG", *args, **kwargs, file=sys.stderr)


def info(*args, **kwargs):
    if (conf.LOG_LEVEL == LogLevel.INFO or conf.LOG_LEVEL == LogLevel.DEBUG) and not conf.TEST_MODE:
        print("INFO", *args, **kwargs, file=sys.stderr)


def warn(*args, **kwargs):
    if (
        conf.LOG_LEVEL == LogLevel.INFO
        or conf.LOG_LEVEL == LogLevel.DEBUG
        or conf.LOG_LEVEL == LogLevel.WARN
    ):
        warnings.warn(*args, **kwargs)


def error(*args, **kwargs):
    print("ERROR", *args, **kwargs, file=sys.stderr)


__all__ = ["debug", "info", "warn", "error"]

# Automatically added by katversion
__version__ = '1.8.3'



from .LSTInterval import LSTInterval, normalize_interval
from .LSTIntervalType import LSTIntervalType
from .LSTIndex import LSTIndex

__all__ = ["LSTIndex", "LSTInterval", "LSTIntervalType", "normalize_interval"]

# Automatically added by katversion
__version__ = '1.8.3'



"""
lstpressure.lst.LST
"""

import pandas as pd
from pandas import DataFrame
from datetime import datetime
from typing import Optional, Callable
from ..observation import Observation
from ..observable import Observable
from .helpers import calculate_observations, calculate_observables
from collections import defaultdict
from io import StringIO
from lstpressure.perf import track_total_runtime, decorate_all
from lstpressure.conf import Conf, LogLevel, LocationProviderType
from lstpressure.logger import info, warn
import json

conf = Conf()


@decorate_all(track_total_runtime)
class LST:
    """
    Wrap the LSTCalendar/Observation API for use with SARAO OPT
    CSV downloads.

    Parameters
    ----------
    input : DataFrame | str | list[list[str]]
        The input data source, which can be a DataFrame, a file path to a CSV file, or a list of lists (rows).
    input_filter : Optional[Callable[[list[list[str]]], bool]]
        A filter function to apply to the input data during instantiation. Defaults to None.
    observation_filter : Optional[Callable[[list[Observation]], bool]]
        A filter function to apply to the observations data after filtering the input data. Defaults to None.
    calendar_start : Optional[str | datetime]
        The start date for the LST calendar. Defaults to the current date.
    calendar_end : Optional[str | datetime]
        The end date for the LST calendar. Defaults to the value of calendar_start.

    Attributes
    ----------
    df : DataFrame
        The filtered DataFrame based on the input data and filters.
    input : str | list[list[str]]
        The input data source, which can be a file path to a CSV file, a DataFrame, or a list of lists (rows).
    calendar_start : str | datetime
        The start date for the LST calendar.
    calendar_end : str | datetime
        The end date for the LST calendar.
    observation_filter : Callable[[list[Observation]], bool]
        The filter function applied to the observations data.
    observations : list[Observation]
        A list of Observation objects based on the filtered data.
    observables : list[Observable]
        A list of Observable objects calculated based on the observations.

    Methods
    -------
    write_to_csv(output: str) -> None:
        Write the observables data to a CSV file.

    """

    def __init__(
        self,
        input: DataFrame | str | list[list[str]],
        input_filter: Optional[Callable[[list[str]], bool]] = None,
        observation_filter: Optional[Callable[[Observation], bool]] = None,
        calendar_start: Optional[str | datetime] = None,
        calendar_end: Optional[str | datetime] = None,
        latitude: Optional[str | float] = conf.LATITUDE,
        longitude: Optional[str | float] = conf.LONGITUDE,
        provider: Optional[LocationProviderType] = None,
    ) -> None:
        """
        Initialize an LST object.

        Parameters
        ----------
        input : DataFrame | str | list[list[str]]
            The input data source, which can be a DataFrame, a file path to a CSV file, or a list of lists (rows).
        input_filter : Optional[Callable[[list[str]], bool]]
            A filter function to apply to the input data during instantiation. Defaults to None.
        observation_filter : Optional[Callable[[Observation], bool]]
            A filter function to apply to the observations data after filtering the input data. Defaults to None.
        calendar_start : Optional[str | datetime]
            The start date for the LST calendar. Defaults to the current date.
        calendar_end : Optional[str | datetime]
            The end date for the LST calendar. Defaults to the value of calendar_start.
        latitude : Optional[str | float]
            Latitude for the location. Defaults to -30:42:39.8 (for ASTRAL provider).
        longitude : Optional[str | float]
            Longitude for the location. Defaults to 21:26:38.0 (for ASTRAL provider).

        Raises
        ------
        TypeError
            If the input is not a valid type (DataFrame, str, or list of lists).
        """
        self._input = input
        self._calendar_start = (
            calendar_start if calendar_start else datetime.now().strftime("%Y%m%d")
        )
        self._calendar_end = calendar_end or calendar_start

        self._latitude = latitude
        self._longitude = longitude
        self._provider = provider

        if input_filter and observation_filter:
            warn(
                "Both 'input_filter' and 'observation_filter' are set. Be aware that 'input_filter' "
                "is applied during instantiation with CSV row data, while 'observation_filter' is applied "
                "afterwards to the already filtered data. Future versions of this API might deprecate "
                "'observation_filter' to streamline data processing.",
                category=UserWarning,
            )

        self._observation_filter = (
            observation_filter if observation_filter else lambda _: True
        )
        input_filter = input_filter if input_filter else lambda _: True

        # Process input based on its type
        if isinstance(input, str):
            # Input is a file path, read CSV and filter rows
            df = pd.read_csv(input)
        elif isinstance(input, list):
            # Input is a list of lists, the first list is treated as header
            header, *data = input
            df = pd.DataFrame(data, columns=header)
        elif isinstance(input, DataFrame):
            # Input is already a DataFrame, make a copy
            df = input.copy()
        else:
            raise TypeError(
                "Input must be a path to a CSV file, a DataFrame, or a list of lists (rows)."
            )

        # Apply the filter to the DataFrame
        self._df = df[df.apply(input_filter, axis=1)]

    @property
    def provider(self) -> LocationProviderType:
        return self._provider

    @property
    def df(self) -> DataFrame:
        """
        The filtered DataFrame based on the input data and filters.

        Returns
        -------
        DataFrame
            The filtered DataFrame.
        """
        return self._df

    @property
    def input(self) -> str | list[list[str]]:
        """
        The input data source, which can be a file path to a CSV file, a DataFrame, or a list of lists (rows).

        Returns
        -------
        str | list[list[str]]
            The input data source.
        """
        return self._input

    @property
    def calendar_start(self) -> str | datetime:
        """
        The start date for the LST calendar.

        Returns
        -------
        str | datetime
            The start date.
        """
        return self._calendar_start

    @property
    def calendar_end(self):
        """
        The end date for the LST calendar.

        Returns
        -------
        str | datetime
            The end date.
        """
        return self._calendar_end

    @property
    def observation_filter(self) -> Callable[[list[Observation]], bool]:
        """
        The filter function applied to the observations data.

        Returns
        -------
        Callable[[list[Observation]], bool]
            The observation filter function.
        """
        return self._observation_filter

    @property
    def latitude(self) -> str | float:
        return self._latitude

    @property
    def longitude(self) -> str | float:
        return self._longitude

    @property
    def observations(self) -> list[Observation]:
        """
        A list of Observation objects based on the filtered data.

        Returns
        -------
        list[Observation]
            A list of Observation objects.
        """
        return calculate_observations(self.df, self.observation_filter)

    @property
    def observables(self) -> list[Observable]:
        """
        A list of Observable objects calculated based on the observations.

        Returns
        -------
        list[Observable]
            A list of Observable objects.
        """
        return calculate_observables(
            self.calendar_start,
            self.calendar_end,
            self.observations,
            self.latitude,
            self.longitude,
            provider=self.provider,
        )

    def to_csv(self) -> DataFrame:
        if conf.LOG_LEVEL != LogLevel.WARN:
            info("Creating CSV...")

        data = [o.to_tuple() for o in self.observables]

        # Use a defaultdict to group data by date
        grouped_data = defaultdict(list)
        for record in data:
            (
                id,
                proposal_id,
                date,
                constraint,
                duration,
                tags,
                preferred_dates,
                avoid_dates,
                lst_window_start,
                lst_window_end,
                lst_interval_start,
                lst_interval_end,
                utc_interval_start,
                utc_interval_end,
            ) = record
            grouped_data[date].append(
                [
                    id,
                    proposal_id,
                    constraint,
                    duration,
                    tags,
                    preferred_dates,
                    avoid_dates,
                    lst_window_start,
                    lst_window_end,
                    lst_interval_start,
                    lst_interval_end,
                    utc_interval_start,
                    utc_interval_end,
                ]
            )

        # Create a list of dictionaries to construct the DataFrame
        data_list = [
            {
                "Date": date,
                "Observation ID": id,
                "Proposal ID": proposal_id,
                "Interval name": constraint,
                "Duration": duration,
                "Observation window start (LST)": lst_window_start,
                "Observation window end (LST)": lst_window_end,
                "Interval start (LST)": lst_interval_start,
                "Interval end (LST)": lst_interval_end,
                "Interval start (UTC)": utc_interval_start,
                "Interval end (UTC)": utc_interval_end,
                "Tags": json.dumps(tags) if len(tags) else None,
                "Preferred dates": json.dumps(preferred_dates)
                if len(preferred_dates)
                else None,
                "Avoid dates": json.dumps(avoid_dates) if len(avoid_dates) else None,
            }
            for date, id_value in grouped_data.items()
            for id, proposal_id, constraint, duration, tags, preferred_dates, avoid_dates, lst_window_start, lst_window_end, lst_interval_start, lst_interval_end, utc_interval_start, utc_interval_end in id_value
        ]

        # Construct the DataFrame
        df = pd.DataFrame(data_list)
        return df

    def to_csv_buffer(self) -> StringIO:
        df = self.to_csv()
        buffer = StringIO()
        df.to_csv(
            buffer, sep=",", quotechar='"', quoting=1, index=False, encoding="utf-8"
        )
        return buffer

    def to_csv_string(self) -> str:
        return self.to_csv_buffer().getvalue()

    def to_csv_file(self, output: str) -> None:
        df = self.to_csv()
        df.to_csv(
            output, sep=",", quotechar='"', quoting=1, index=False, encoding="utf-8"
        )



from datetime import datetime
from typing import Optional
from ...observable import Observable
from ...observation import Observation
from ...lstcalendar import LSTCalendar
from lstpressure.perf import track_total_runtime
from lstpressure.conf import LocationProviderType
from typing import Optional


@track_total_runtime
def calculate_observables(
    cal_start: str | datetime,
    cal_end: str | datetime,
    observations: list[Observation],
    latitude: Optional[str | float] = None,
    longitude: Optional[str | float] = None,
    provider: Optional[LocationProviderType] = None,
) -> list[Observable]:
    return sorted(
        LSTCalendar(
            cal_start,
            cal_end,
            observations=observations,
            latitude=latitude,
            longitude=longitude,
            provider=provider,
        ).observables()
    )



"""
lstpressure.time_conversions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Module containing time conversion utilities, including conversion from UTC to Local Sidereal Time (LST) 
and conversion from a time string to its decimal hour representation.
"""

from astropy.time import Time
from astropy import units as u, coordinates
from functools import lru_cache
from astropy.utils import iers
from datetime import datetime
from .normalize_coordinates import normalize_coordinates
from .normalize_date import normalize_datetime
from lstpressure.perf import track_total_runtime


# iers.conf.auto_download = False
# iers.conf.auto_max_age = None
# iers.conf.iers_degraded_accuracy = "warn"

# Constants
LST_DAY_DEC = 23.9344696
"""Length of a sidereal day in decimal hours."""


@track_total_runtime
@lru_cache(maxsize=None)
def utc_to_lst(
    iso_date: str | datetime,
    lat: str | float = None,
    long: str | float = None,
    address: str = None,
) -> float:
    """
    Convert a given UTC datetime to Local Sidereal Time (LST).

    Parameters:
    -----------
    iso_date : str | datetime
        An ISO 8601 date string or datetime object representing UTC time.
    lat : str | float
        Latitude coordinate, can be a string (e.g., "52d40m") or float.
    long : str | float
        Longitude coordinate, can be a string (e.g., "4d55m") or float.
    address : str
        Address that is resolved to lat/long (instead of passing lat/long) via OpenStreetMap API

    Returns:
    --------
    float
        The Local Sidereal Time in decimal hours.

    Example:
    --------
    >>> utc_to_lst("2023-10-26T12:00:00", "52d40m", "4d55m")
    14.5567
    """
    if address:
        location = coordinates.EarthLocation.of_address(address)
        longitude = location.lon
        latitude = location.lat
        long = longitude.value
        lat = latitude.value

    lat, long = normalize_coordinates(lat, long)
    location = coordinates.EarthLocation(long, lat)
    t = Time(val=normalize_datetime(iso_date), format="datetime", location=location)
    lst = t.sidereal_time(kind="apparent", model="IAU2006A")
    return lst.to_value()


@track_total_runtime
@lru_cache(maxsize=None)
def normalize_time_to_decimal(time: str | float) -> float:
    """
    Convert a time string of format hours:min:sec to decimal hours.

    Parameters:
    -----------
    time : str
        A time string formatted as "hours:min:sec".

    Returns:
    --------
    float
        The time represented in decimal hours.

    Raises:
    -------
    ValueError
        If the input time is not a string or if it doesn't have the expected format.

    Example:
    --------
    >>> time_to_decimal("2:30:0")
    2.5
    """
    if isinstance(time, float):
        return time

    if not isinstance(time, str):
        raise ValueError("Input should be a string in format hours:min:sec")

    # Split the time string by colon to get hours, minutes, and seconds
    components = time.split(":")

    if len(components) != 3:
        raise ValueError("Time string should have format hours:min:sec")

    hours, minutes, seconds = map(int, components)

    # Convert time components to decimal hours
    decimal_hours = hours + (minutes / 60) + (seconds / 3600)

    return decimal_hours



import pytest
from lstpressure.utils import utc_to_lst, normalize_time_to_decimal, normalize_coordinates


@pytest.mark.parametrize("input,expected", [("6:26:20", 6.438888888888889)])
def test_normalize_time_to_decimal(input, expected):
    result = normalize_time_to_decimal(input)
    assert result == expected


# https://www.localsiderealtime.com/


@pytest.mark.parametrize(
    "utc,latitude,longitude,expected",
    [
        ("2023-10-26 09:21:50", "-30:42:39.8", "21:26:38.0", "13:05:36"),
        ("2023-10-26 08:26:20", "-30:42:39.8", "21:26:38.0", 12.165872678174422),
        ("2023-10-26 09:21:50", -30.7111, 21.4439, "13:05:36"),
    ],
)
def test_utc_to_lst(utc, latitude, longitude, expected):
    result = utc_to_lst(utc, *normalize_coordinates(latitude, longitude))
    assert round(normalize_time_to_decimal(result), 2) == round(
        normalize_time_to_decimal(expected), 2
    )



"""
lstpressure.lstcalendar.LSTCalendar
"""
from datetime import timedelta, datetime
from typing import Dict, Optional, Self
from ..observation import Observation
from ..observable import Observable
from .LSTCalendarDate import LSTCalendarDate
from ..lstindex import LSTIndex, LSTInterval
from ..utils import (
    normalize_yyyymmdd_to_datetime,
    normalize_coordinates,
)
from lstpressure.perf import decorate_all, track_total_runtime
from lstpressure.logger import info
from lstpressure.conf import Conf, LocationProviderType

conf = Conf()


@decorate_all(track_total_runtime)
class LSTCalendar:
    """
    Calendar tailored for LST (Local Sidereal Time) related interval lookups.

    Attributes
    ----------
    start : str | datetime
        The beginning of the date range for the calendar.
    end : Optional[str | datetime]
        The conclusion of the date range for the calendar. Defaults to None.
    latitude : float
        The geographic latitude in decimal degrees. Defaults to 0.
    longitude : float
        The geographic longitude in decimal degrees. Defaults to 0.
    interval_index : LSTIndex
        An index to manage intervals efficiently.
    observations_index : LSTIndex
        An index to manage observations efficiently.
    dates : list[LSTCalendarDate]
        A list containing dates and corresponding sun statistics within the range.

    Methods
    -------
    __init__(start, end, latitude=0, longitude=0, observations=[])
        Initialize the LSTCalendar object.
    __enter__()
        Enter context management.
    __exit__(exc_type, exc_val, exc_tb)
        Exit context management.
    load_observations(observations: list[Observation]) -> LSTCalendar
        Load a list of observations and insert their intervals into the observation index.
    observables(observations: Optional[list[Observation]] = None) -> list[Observable]
        Retrieve observable observations for the dates maintained by this calendar.

    """

    def __enter__(self):
        """
        Enter context management.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit context management.
        """
        pass

    def __init__(
        self,
        start: str | datetime,
        end: Optional[str | datetime] = None,
        observations: list[Observation] = [],
        latitude: Optional[str | float] = conf.LATITUDE,
        longitude: Optional[str | float] = conf.LONGITUDE,
        provider: Optional[LocationProviderType] = None,
    ):
        """
        Initialize the LSTCalendar object.

        Parameters
        ----------
        start : str | datetime
            Start date of the calendar range.
        end : (Optional) str | datetime
            End date of the calendar range. Defaults to start.
        observations : list[Observation], optional
            list of observation instances. Called load_observations method under the hood.
        latitude : str | float, optional
            Latitude for the location. Defaults to -30:42:39.8 (for ASTRAL provider).
        longitude : str | float, optional
            Longitude for the location. Defaults to 21:26:38.0 (for ASTRAL provider).

        Raises
        ------
        ValueError
            If the start date is after the end date.
        """
        info(f"Initializing calendar ({conf.LOC_PROVIDER.name} location provider)...")
        start = normalize_yyyymmdd_to_datetime(start)
        end = start if not end else normalize_yyyymmdd_to_datetime(end)

        if start > end:
            raise ValueError("start day should be <= end day")

        if latitude and longitude:
            latitude, longitude = normalize_coordinates(latitude, longitude)
        self.latitude = latitude
        self.longitude = longitude
        self._interval_index = LSTIndex()
        self._observations_index = LSTIndex()
        self._dates = [
            LSTCalendarDate(start + timedelta(days=d), self, provider=provider)
            for d in range(0, (end - start).days + 1)
        ]
        self.load_observations(observations)

    @property
    def interval_index(self) -> LSTIndex:
        """
        Property to retrieve the interval index.

        Returns
        -------
        LSTIndex
            An index to manage intervals efficiently.
        """
        return self._interval_index

    @property
    def observations_index(self) -> LSTIndex:
        """
        Property to retrieve the observations index.

        Returns
        -------
        LSTIndex
            An index to manage observations efficiently.
        """
        return self._observations_index

    @property
    def dates(self) -> list[LSTCalendarDate]:
        """
        Property to retrieve the list of dates.

        Returns
        -------
        list[LSTCalendarDate]
            A list containing dates and corresponding sun statistics within the range.
        """
        return self._dates

    def load_observations(self, observations: list[Observation]) -> Self:
        """
        Load a list of observations, set the calendar attribute for each observation,
        and insert their intervals into the observation index.

        Observations must have unique IDs.

        Parameters
        ----------
        observations : list[Observation]
            A list of observation instances.

        Returns
        -------
        LSTCalendar
            The LSTCalendar object.
        """
        self._observations = []
        self._observations_index = LSTIndex()
        _primary_keys = set()
        for observation in observations:
            # Create a composite key using both id and utc_constraints
            composite_key = (
                observation.id,
                "".join([c.name for c in observation.utc_constraints]),
            )

            # Check if the composite key is already in the set
            if composite_key in _primary_keys:
                raise ValueError(
                    f"Duplicate composite key detected: ID '{observation.id}' with UTC constraints '{observation.utc_constraints}'. "
                    "Each observation must have a unique combination of ID and UTC constraints to be loaded into the calendar."
                )

            # Add the composite key to the set for future checks
            _primary_keys.add(composite_key)

            # Assign the calendar to the observation and append it to the calendar's list of observations
            observation.calendar = self
            self._observations.append(observation)

            # Insert intervals from the observation into the calendar's observation index
            for interval in observation.intervals:
                self.observations_index.insert(interval.interval)
        return self

    @property
    def observations(self) -> list[Dict[str, LSTInterval | Observation]]:
        """
        Property to retrieve the list of observations.

        Returns
        -------
        list[Dict[str, LSTInterval | Observation]]
            A list of observations.
        """
        return self._observations

    @observations.setter
    def observations(self, observations: list[Observation]) -> None:
        """
        Setter for the observations property. It serves as an alias for the load_observations method.

        Parameters
        ----------
        observations : list[Observation]
            A list of observation instances.

        Returns
        -------
        None
        """
        self.load_observations(observations)

    def observables(
        self, observations: Optional[list[Observation]] = None
    ) -> list[Observable]:
        """
        Retrieve observable observations for the dates maintained by this calendar.

        This method processes a list of dates within the calendar and aggregates their observable observations into observation windows. If a list of observations is provided, it replaces the current observations of the calendar with the new list.

        Parameters
        ----------
        observations : Optional[list[Observation]], optional
            A list of Observation objects to process, which is optional. If provided, it overwrites any existing observations in the calendar.

        Raises
        ------
        TypeError
            If no observations have been loaded into the calendar and none are provided.

        Returns
        -------
        list[Observable]
            A list of aggregated observation windows, each representing observable observations for a specific date in the calendar.

        Example usage:

        .. code-block:: python

            calendar = Calendar()
            calendar.load_dates(dates)
            try:
                observation_windows = calendar.observable_observations()
            except TypeError as e:
                print(e)

        """
        # If observations provided as an argument, assume the intent is to overwrite
        if observations:
            self.load_observations(observations)

        if not self.observations:
            raise TypeError(
                "No observations have been loaded on this calendar, and no observations have been provided as an argument"
            )

        results = []
        for dt in self.dates:
            results = results + dt.observables()

        return results



from .AstroUtils import AstroUtil
from .SolarSystemUtils import SolarSystem


class DateTimeUtil:
    # You can implement the local_sidereal function based on your specific need

    @staticmethod
    def local_sidereal(jd):
        # Assuming g_sidereal is implemented elsewhere
        res = SolarSystem.g_sidereal(jd)  # You need to implement g_sidereal

        # Adding longitude and wrapping to 360 degrees
        res += AstroUtil.MKAT_POSITION["longitude"]
        return AstroUtil.rev360(res)

    @staticmethod
    def jd(date):
        yr = date.year
        mon = date.month
        day = date.day
        hr = date.hour
        min = date.minute
        sec = date.second
        fday = day + ((3600.0 * hr + 60.0 * min + sec) / 86400.0)

        if mon <= 2:
            yr -= 1
            mon += 12

        if yr > 1582:  # 1582 October 15
            A = yr // 100
            B = 2 - A + A // 4
        else:
            A = 0
            B = 0

        if yr < 0:
            C = int((365.25 * yr) - 0.75)
        else:
            C = int(365.25 * yr)

        D = int(30.6001 * (mon + 1))

        jd = B + C + D + fday + 1720994.5

        return jd

    @staticmethod
    def lst2gst(lst, longitude):
        return AstroUtil.rev24(lst - (longitude / 15))

    @staticmethod
    def gst2ut(jd, gst):
        S = jd - 2451545.0
        T = S / 36525.0
        T0 = AstroUtil.rev24(6.697374558 + (2400.051336 * T))
        ut = AstroUtil.rev24(gst - T0) * 0.9972695663
        return ut

    @staticmethod
    def lst2ut(lst, longitude, date):
        gst = DateTimeUtil.lst2gst(lst, longitude)
        jd = DateTimeUtil.jd(date)
        return DateTimeUtil.gst2ut(jd, gst)



import math


class AstroUtil:
    RTD = 180.0 / math.pi
    DTR = math.pi / 180

    MKAT_POSITION = {
        "longitude": 21.443888888888889,
        "latitude": -30.7110555555556,
        "limit": 15,
        "tz": +2,
    }

    @staticmethod
    def sind(angle):
        return math.sin(math.radians(angle))

    @staticmethod
    def asind(angle):
        return AstroUtil.RTD * math.asin(angle)

    @staticmethod
    def cosd(angle):
        return math.cos(math.radians(angle))

    @staticmethod
    def acosd(angle):
        return math.degrees(math.acos(angle))

    @staticmethod
    def atan2d(y, x):
        return math.degrees(math.atan2(y, x))

    @staticmethod
    def tand(angle):
        return math.tan(math.radians(angle))

    @staticmethod
    def rev360(angle):
        return angle % 360

    @staticmethod
    def rev24(hour):
        return hour - (hour // 24.0) * 24.0

    @staticmethod
    def hms(x):
        hms_list = x.split(":")
        hr = int(hms_list[0])
        min = int(hms_list[1])
        sec = float(hms_list[2])
        decimal = hr + min / 60.0 + sec / 3600.0
        return decimal



from lstpressure.perf import track_total_runtime


@track_total_runtime
def normalize_interval(start, end, days=1):
    if end < start:
        end += 24
    end += 24 * (days - 1)
    return (start, end)



from setuptools import setup, find_packages
import os

def parse_requirements():
    """Parses the requirements.txt file and returns a list of dependencies."""
    requirements = []
    requirements_file = "requirements.txt"

    if os.path.isfile(requirements_file):
        with open(requirements_file, "r", encoding="utf-8") as f:
            for line in f:
                # Ignore comments and empty lines
                line = line.strip()
                if line and not line.startswith("#"):
                    requirements.append(line)
    return requirements


def get_long_description():
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()
        return long_description

requirements = parse_requirements()
print("requirements", requirements)

setup(
    name="lst-pressure",
    setup_requires=["katversion"],
    use_katversion=True,
    description='Determine periods of "LST pressure" by querying for intersections between LST/Solar intervals',
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/ska-sa/lst-pressure",
    author="Zach Smith",
    author_email="zsmith@sarao.ac.za",
    license="Apache 2.0",
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.12",
    ],
    package_dir={},
    packages=find_packages(where="./"),
    python_requires=">=3.12",
    install_requires=requirements,
)



"""
lstcalendar
"""
from .LSTCalendar import LSTCalendar
from .LSTCalendarDate import LSTCalendarDate

__all__ = [
    "LSTCalendar",
    "LSTCalendarDate",
]

# Automatically added by katversion
__version__ = '1.8.3'



import pytest
from lstpressure.lstindex import LSTInterval, LSTIntervalType
from intervaltree import Interval
from datetime import datetime


@pytest.mark.parametrize(
    "start, end, dt, interval_type",
    [
        (
            1,
            2,
            "20230219",
            LSTIntervalType.NIGHT,
        ),
        (
            1,
            2,
            "20230219",
            None,
        ),
    ],
)
def test_LSTInterval(start, end, dt, interval_type):
    interval = LSTInterval(start, end, None, None, None, dt, interval_type)
    _interval = interval.interval
    assert isinstance(_interval, Interval)
    assert isinstance(_interval[2], LSTInterval)
    assert interval.start == _interval[0]
    assert interval.end == _interval[1]
    assert interval.type == interval_type
    assert interval.dt == datetime.strptime(dt, "%Y%m%d").date()



import pytest
from lstpressure import LSTCalendar, LSTIntervalType as I
from lstpressure.observation import Observation

OBSERVATIONS = [
    Observation(
        o.get("id"),
        o.get("lst_window_start"),
        o.get("lst_window_end"),
        o.get("utc_constraints"),
        o.get("duration"),
    )
    for o in [
        {
            "id": "O1",
            "lst_window_start": 6.5,
            "lst_window_end": 10.2,
            "utc_constraints": [I.ALL_DAY],
            "duration": 1,
        },
        {
            "id": "O2",
            "lst_window_start": 8.5,
            "lst_window_end": 10.2,
            "utc_constraints": [I.NIGHT],
            "duration": 1,
        },
        {
            "id": "O3",
            "lst_window_start": 21.5,
            "lst_window_end": 20,
            "utc_constraints": [I.NIGHT],
            "duration": 0.5,
        },
        {
            "id": "O4",
            "lst_window_start": 12,
            "lst_window_end": 22,
            "utc_constraints": [I.NIGHT],
            "duration": 1.5,
        },
        {
            "id": "O5",
            "lst_window_start": 9,
            "lst_window_end": 10,
            "utc_constraints": [I.SUNRISE_SUNSET],
            "duration": 0.5,
        },
    ]
]


@pytest.mark.parametrize(
    "start, end, expected",
    [
        ("20230404", "20230404", ["20230404"]),
        ("20230404", "20230405", ["20230404", "20230405"]),
        ("20220101", "20220105", ["20220101", "20220102", "20220103", "20220104", "20220105"]),
        (
            "20231025",
            "20231031",
            ["20231025", "20231026", "20231027", "20231028", "20231029", "20231030", "20231031"],
        ),
    ],
)
def test_Calendar(start, end, expected):
    """
    The calendar should convert start/end params into the correct range
    """
    assert expected == [d.dt.strftime("%Y%m%d") for d in LSTCalendar(start, end)._dates]


# Invalid start/end should NOT work
@pytest.mark.parametrize(
    "start, end",
    [("invalidStart", "20220105"), ("20220101", "invalidEnd"), ("20220105", "20220101")],
)
def test_calendar_raises_exception_for_invalid_dates(start, end):
    with pytest.raises(ValueError):
        LSTCalendar(start, end)


# The observations property should return all the observations
@pytest.mark.parametrize(
    "dt_range", [(["20231101"]), (["20231101", "20231101"]), (["20231101", "20231201"])]
)
def test_self_observations(dt_range):
    start = dt_range[0]
    end = dt_range[1] if len(dt_range) == 2 else start
    calendar = LSTCalendar(start, end, observations=OBSERVATIONS)
    assert len(calendar.observations) == len(OBSERVATIONS)



"""
normalize_date.py
"""
from functools import lru_cache
from datetime import date, datetime
from lstpressure.perf import track_total_runtime


@track_total_runtime
@lru_cache(maxsize=None)
def normalize_yyyymmdd_to_datetime(yyyymmdd) -> date:
    """
    normalize_date
    In many cases it's useful to create a date from a yyyymmdd string
    This helper function is to allow inputting either a string or date
    object and getting back a date object at 00:00 of that day
    """
    if isinstance(yyyymmdd, str):
        year, month, day = int(yyyymmdd[:4]), int(yyyymmdd[4:6]), int(yyyymmdd[6:])
    else:
        year, month, day = yyyymmdd.year, yyyymmdd.month, yyyymmdd.day
    return date(year, month, day)


@track_total_runtime
@lru_cache(maxsize=None)
def normalize_datetime(dt_input) -> datetime:
    """
    normalize_datetime
    Normalizes a datetime object (either a string or a date) to a datetime object

    dt_input("2023-12-20 00:00:00"): <class 'datetime.datetime'> 2023-12-30 00:00:00
    dt_input(datetime(2023, 12, 20, 0, 0, 0)): <class 'datetime.datetime'> 2023-12-30 00:00:00
    """
    if isinstance(dt_input, str):
        return datetime.fromisoformat(dt_input)
    return dt_input
